import numpy as np
from typing import Callable, Union
import tqdm
from pyuncertainnumber.propagation.epistemic_uncertainty.cartesian_product import cartesian
from pyuncertainnumber.propagation.epistemic_uncertainty.extreme_point_func import extreme_pointX
from pyuncertainnumber.propagation.epistemic_uncertainty.extremepoints import extremepoints_method
from pyuncertainnumber.propagation.utils import Propagation_results, condense_bounds

# TODO add tail concentrating algorithms.
# TODO add x valus for min and max


def second_order_propagation_method(x: list, f: Callable = None,
                                    results: Propagation_results = None,
                                    method='endpoints',
                                    n_disc: Union[int, np.ndarray] = 10,
                                    condensation: Union[float,
                                                        np.ndarray] = None,
                                    tOp: Union[float, np.ndarray] = 0.999,
                                    bOt: Union[float, np.ndarray] = 0.001,
                                    save_raw_data='no') -> Propagation_results:  # Specify return type
    """
    args:
        - x (list): A list of `UncertainNumber` objects representing the uncertain inputs.
        - f (Callable): The function to evaluate.
        - results (Propagation_results, optional): An object to store propagation results.
                                    Defaults to None, in which case a new
                                    `Propagation_results` object is created.
        - method (str, optional): The method used to estimate the bounds of each combination 
                            of focal elements. Can be either 'endpoints' or 'extremepoints'. 
                            Defaults to 'endpoints'.
        - n_disc (Union[int, np.ndarray], optional): The number of discretization points 
                                    for each uncertain input. If an integer is provided,
                                    it is used for all inputs. If a NumPy array is provided,
                                    each element specifies the number of discretization 
                                    points for the corresponding input. 
                                    Defaults to 10.
        - condensation (Union[float, np.ndarray], optional): A parameter or array of parameters 
                                    to control the condensation of the output p-boxes. 
                                    Defaults to None.
        - tOp (Union[float, np.ndarray], optional): Upper threshold or array of thresholds for 
                                    discretization. 
                                    Defaults to 0.999.
        - bOt (Union[float, np.ndarray], optional): Lower threshold or array of thresholds for 
                                    discretization. 
                                    Defaults to 0.001.
        - save_raw_data (str, optional): Whether to save raw data ('yes' or 'no'). 
                                   Defaults to 'no'.

    signature:
        second_order_propagation_method(x: list, f: Callable, results: Propagation_results = None, ...) -> Propagation_results

    notes:
        - Performs second-order uncertainty propagation for mixed uncertain numbers.
        - The function handles different types of uncertain numbers (distributions, intervals, 
          p-boxes) and discretizes them accordingly.
        - It generates combinations of focal elements from the discretized uncertain inputs.
        - For the 'endpoints' method, it evaluates the function at all combinations of endpoints 
          of the focal elements.
        - For the 'extremepoints' method, it uses the `extremepoints_method` to determine the 
          signs of the partial derivatives and evaluates the function at the extreme points.
        - The output p-boxes are constructed by considering the minimum and maximum values obtained 
          from the function evaluations.
        - The `condensation` parameter can be used to reduce the number of intervals in the output 
          p-boxes. 

    returns:
        Propagation_results:  A `Propagation_results` object containing:
            - raw_data (dict): Dictionary containing raw data shared across output(s):
                    - x (np.ndarray): Input values.
                    - f (np.ndarray): Output values.
                    - min (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the minimum of that output.
                    - max (np.ndarray): Array of dictionaries, one for each output,
                              containing 'f' for the maximum of that output.
                    - bounds (np.ndarray): 2D array of lower and upper bounds for each output.

    example:
        from pyuncertainnumber import UncertainNumber

        def Fun(x):

            input1= x[0]
            input2=x[1]
            input3=x[2]
            input4=x[3]
            input5=x[4]

            output1 = input1 + input2 + input3 + input4 + input5
            output2 = input1 * input2 * input3 * input4 * input5

            return np.array([output1, output2])

        means = np.array([1, 2, 3, 4, 5])
        stds = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

        x = [
            UncertainNumber(essence = 'distribution', distribution_parameters= ["gaussian",[means[0], stds[0]]]),

            UncertainNumber(essence = 'interval', bounds= [means[1]-2* stds[1], means[1]+2* stds[1]]),
            UncertainNumber(essence = 'interval', bounds= [means[2]-2* stds[2], means[2]+2* stds[2]]),
            UncertainNumber(essence = 'interval', bounds= [means[3]-2* stds[3], means[3]+2* stds[3]]),
            UncertainNumber(essence = 'interval', bounds= [means[4]-2* stds[4], means[4]+2* stds[4]])
            ]

        results = second_order_propagation_method(x=x, f=Fun, method = 'endpoints', n_disc= 5)

    """
    d = len(x)  # dimension of uncertain numbers
    results = Propagation_results()
    bounds_x = []
    ranges = np.zeros((2, d))
    u_list = []  # List to store 'u' for each uncertain number

    for i, un in enumerate(x):
        print(f"Processing variable {i + 1} with essence: {un.essence}")

        if un.essence != "interval":
            # perform outward directed discretisation
            if isinstance(n_disc, int):
                nd = n_disc + 1
            else:
                nd = n_disc[i] + 1  # Use the i-th element of n_disc array

            distribution_family = un.distribution_parameters[0]

            if distribution_family == 'triang':
                u = np.linspace(0, 1, nd)
            else:
                if isinstance(tOp, np.ndarray):
                    top = tOp[i]
                else:
                    top = tOp
                if isinstance(bOt, np.ndarray):
                    bot = bOt[i]
                else:
                    bot = bOt
                u = np.linspace(bot, top, nd)  # Use bOt and tOp here

        else:  # un.essence == "interval"
            u = np.array([0.0, 1.0])  # Or adjust as needed for intervals

        u_list.append(u)  # Add 'u' to the list

        # Generate discrete p-boxes
        match un.essence:
            case "distribution":
                # Calculate xl and xr for distributions (adjust as needed)
                # Assuming un.ppf(u) returns a list or array
                temp_xl = un.ppf(u_list[i][:-1])
                # Adjust based on your distribution
                temp_xr = un.ppf(u_list[i][1:])
                # Create a 2D array of bounds
                rang = np.array([temp_xl, temp_xr]).T
                bounds_x.append(rang)
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])

            case "interval":
                # Repeat lower bound for all quantiles
                temp_xl = np.array([un.bounds[0]])
                # Repeat upper bound for all quantiles
                temp_xr = np.array([un.bounds[1]])
                # Create a 2D array of bounds
                rang = np.array([temp_xl, temp_xr]).T
                bounds_x.append(rang)
                ranges[:, i] = un.bounds  # np.array([un.bounds])

            case "pbox":
                temp_xl = un.ppf(u_list[i][:-1])[0]
                temp_xr = un.ppf(u_list[i][1:])[1]
                # Create a 2D array of bounds
                rang = np.array([temp_xl, temp_xr]).T
                bounds_x.append(rang)
                ranges[:, i] = np.array([temp_xl[0], temp_xr[-1]])

            case _:
                raise ValueError(f"Unsupported uncertainty type: {un.essence}")

    # Automatically generate merged_array_index
    bounds_x_index = [np.arange(len(sub_array)) for sub_array in bounds_x]

    # Calculate Cartesian product of indices using your cartesian function
    cartesian_product_indices = cartesian(*bounds_x_index)

    # Generate the final array using the indices
    focal_elements_comb = []
    for indices in cartesian_product_indices:
        temp = []
        for i, index in enumerate(indices):
            temp.append(bounds_x[i][index])
        focal_elements_comb.append(temp)

    focal_elements_comb = np.array(focal_elements_comb, dtype=object)
    all_output = None

    if f is not None:
        # Efficiency upgrade: store repeated evaluations
        inpsList = np.zeros((0, d))
        evalsList = []
        numRun = 0

        match method:
            case "endpoints" | "second_order_endpoints":
                # Pre-allocate the array
                x_combinations = np.empty(
                    (focal_elements_comb.shape[0]*(2**d), d), dtype=float)
                current_index = 0  # Keep track of the current insertion index

                for array in focal_elements_comb:
                    cartesian_product_x = cartesian(*array)
                    # Get the number of combinations from cartesian(*array)
                    num_combinations = cartesian_product_x.shape[0]

                    # Assign the cartesian product to the appropriate slice of x_combinations
                    x_combinations[current_index: current_index +
                                   num_combinations] = cartesian_product_x
                    current_index += num_combinations  # Update the insertion index

                # Initialize all_output as a list to store outputs initially
                all_output_list = []
                evalsList = []
                numRun = 0
                # Initialize inpsList with the correct number of columns
                inpsList = np.empty((0, x_combinations.shape[1]))

                # Wrap the loop with tqdmx_combinations
                for case in tqdm.tqdm(x_combinations, desc="Evaluating focal points"):
                    im = np.where((inpsList == case).all(axis=1))[0]
                    if not im.size:
                        output = f(case)
                        all_output_list.append(output)
                        inpsList = np.vstack([inpsList, case])
                        evalsList.append(output)
                        numRun += 1
                    else:
                        all_output_list.append(evalsList[im[0]])

                # Determine num_outputs AFTER running the function
                try:
                    num_outputs = len(all_output_list[0])
                except TypeError:
                    num_outputs = 1

                # Convert all_output to a 2D NumPy array
                all_output = np.array(all_output_list).reshape(
                    focal_elements_comb.shape[0], (2**d), num_outputs)

                # Calculate min and max for each sublist in all_output
                min_values = np.min(all_output, axis=1)
                max_values = np.max(all_output, axis=1)

                lower_bound = np.zeros((num_outputs, len(min_values)))
                upper_bound = np.zeros((num_outputs, len(max_values)))

                bounds = np.empty((num_outputs, 2, lower_bound.shape[1]))

                for i in range(num_outputs):
                    min_values[:, i] = np.sort(min_values[:, i])
                    max_values[:, i] = np.sort(max_values[:, i])
                    lower_bound[i, :] = min_values[:, i]  # Extract each column
                    upper_bound[i, :] = max_values[:, i]

                    bounds[i, 0, :] = lower_bound[i, :]
                    bounds[i, 1, :] = upper_bound[i, :]

            case "extremepoints" | "second_order_extremepoints":
                # Determine the positive or negative signs for each input

                res = extremepoints_method(ranges.T, f)

                # Determine the number of outputs from the first evaluation
                try:
                    num_outputs = res.raw_data['sign_x'].shape[0]
                except TypeError:
                    num_outputs = 1  # If f returns a single value

                inpsList = np.zeros((0, d))
                evalsList = np.zeros((0, num_outputs))
                all_output_list = []

                # Preallocate all_output_list with explicit loops
                all_output_list = []
                for _ in range(num_outputs):
                    output_for_current_output = []
                    # Changed to focal_elements_comb
                    for _ in range(len(focal_elements_comb)):
                        output_for_current_output.append([None, None])
                    all_output_list.append(output_for_current_output)

                # Iterate over focal_elements_comb
                for i, slice in tqdm.tqdm(enumerate(focal_elements_comb), desc="Evaluating focal points", total=len(focal_elements_comb)):
                    # Changed to 2 for the two extreme points
                    Xsings = np.empty((2, d))
                    # Use the entire sign_x array
                    Xsings[:, :] = extreme_pointX(
                        slice, res.raw_data['sign_x'])

                    for k in range(Xsings.shape[0]):  #
                        c = Xsings[k, :]
                        im = np.where((inpsList == c).all(axis=1))[0]
                        if not im.size:
                            output = f(c)

                            # Store each output in a separate sublist
                            for out in range(num_outputs):
                                # Changed indexing
                                all_output_list[out][i][k] = output[out]

                            inpsList = np.vstack([inpsList, c])

                            # Ensure output is always a NumPy array
                            if not isinstance(output, np.ndarray):
                                output = np.array(output)

                            evalsList = np.vstack([evalsList, output])
                        else:
                            for out in range(num_outputs):
                                # Changed indexing
                                all_output_list[out][i][k] = evalsList[im[0]][out]

                # Reshape all_output based on the actual number of elements per output
                all_output = np.array(all_output_list, dtype=object)
                all_output = np.reshape(
                    all_output, (num_outputs, -1, 2))  # Reshape to 3D

                # Calculate min and max for each sublist in all_output
                min_values = np.min(all_output, axis=2)
                max_values = np.max(all_output, axis=2)

                lower_bound = np.zeros((num_outputs, min_values.shape[1]))
                upper_bound = np.zeros((num_outputs, max_values.shape[1]))

                bounds = np.empty((num_outputs, 2, lower_bound.shape[1]))

                for i in range(num_outputs):
                    lower_bound[i, :] = np.sort(
                        min_values[i, :])  # Extract each column
                    upper_bound[i, :] = np.sort(max_values[i, :])

                    bounds[i, 0, :] = lower_bound[i, :]
                    bounds[i, 1, :] = upper_bound[i, :]

            case _:
                raise ValueError(
                    "Invalid UP method! endpoints_cauchy are under development.")

        if condensation is not None:
            bounds = condense_bounds(bounds, condensation)

        results.raw_data['bounds'] = bounds
        results.raw_data['min'] = np.array([{'f': lower_bound[i, :]} for i in range(
            num_outputs)])  # Initialize as a NumPy array
        results.raw_data['max'] = np.array([{'f': upper_bound[i, :]} for i in range(
            num_outputs)])  # Initialize as a NumPy array

        if save_raw_data == 'yes':
            print('No raw data provided for this method!')
            # results.add_raw_data(f= all_output, x= x_combinations)

    elif save_raw_data == 'yes':  # If f is None and save_raw_data is 'yes'
        results.add_raw_data(f=None, x=x_combinations)

    else:
        print("No function is provided. Select save_raw_data = 'yes' to save the input combinations")

    return results



from __future__ import annotations
from typing import TYPE_CHECKING
from scipy.stats import qmc
import numpy as np

if TYPE_CHECKING:
    from ...pba.intervals import Interval

"""some helper funcs and classes for the epistemic space"""


class EpistemicDomain:
    """Representation of the epistemic space which are indeed bounds of each dimension

    This class provides a set of handy functions to work with epistemic uncertainty in the form of bounds.
    It will be useful for tasks such as propagation or optimization where epistemic uncertainty is involved.

    args:
        vars: a set of Interval variables

    tip:
        a must to use for optimisation tasks where the design bounds can be quickly specified with the ``to_OptBounds`` method.

    example:
        >>> e = EpistemicDomain(xe1, xe2)
        >>> # convert the epistemic space to bounds for the optimizer
        >>> e.to_OptBounds()
        >>> # perform lhs sampling on the epistemic space
        >>> sample = e.lhs_sampling(1000)
    """

    def __init__(self, *vars: Interval):
        from ...pba.intervals.intervalOperators import make_vec_interval

        self.vec_interval = make_vec_interval(vars)

    def lhs_sampling(self, n_samples: int):
        """perform lhs sampling on the epistemic space"""
        Xc_sampler = qmc.LatinHypercube(d=len(self.vec_interval))
        l_bounds = self.vec_interval.lo
        u_bounds = self.vec_interval.hi

        base_sample = Xc_sampler.random(n=n_samples)
        return qmc.scale(base_sample, l_bounds, u_bounds)

    def lhs_plus_endpoints(self, n_samples: int):
        """perform lhs sampling on the epistemic space and add endpoints"""
        sample = self.lhs_sampling(n_samples)
        endpoints = self.to_OptBounds().T
        combined_sample = np.vstack((sample, endpoints))
        return combined_sample

    def bound_rep(self):
        """return the bounds of the epistemic space"""
        pass

    def to_OptBounds(self) -> np.ndarray:
        """convert the epistemic space to bounds for the optimizer"""
        return self.vec_interval.to_numpy()

    # def to_BObounds(self) -> dict:
    #     """convert the epistemic space to bounds for the Bayesian optimizer"""

    #     new_dict = self.__dict__.copy()
    #     new_dict = {k: tuple(v) for k, v in new_dict.items()}
    #     return new_dict



import numpy as np


def cartesian(*arrays):
    """Computes the Cartesian product of multiple input arrays

    args:
       - *arrays: Variable number of np.arrays representing the sets of values for each dimension.


    signature:
       - cartesian(*x:np.array)  -> np.ndarray

    note:
       - The data type of the output array is determined based on the input arrays to ensure compatibility.

    return:
        - darray: A NumPy array where each row represents one combination from the Cartesian product.
                  The number of columns equals the number of input arrays.

    example:
        >>> x = np.array([1, 2], [3, 4], [5, 6])
        >>> y = cartesian(*x)
        >>> # Output:
        >>> # array([[1, 3, 5],
        >>> #        [1, 3, 6],
        >>> #        [1, 4, 5],
        >>> #        [1, 4, 6],
        >>> #        [2, 3, 5],
        >>> #        [2, 3, 6],
        >>> #        [2, 4, 5],
        >>> #        [2, 4, 6]])

    """
    la = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [la], dtype=dtype)
    for i, a in enumerate(np.ix_(*arrays)):
        arr[..., i] = a
    return arr.reshape(-1, la)


# def cartesian(sides): to be deleted perhaps
#   """
#   Generates the Cartesian product of the input sets.

#   Args:
#     sides: A NumPy array where each row represents a set of values.

#   Returns:
#     A NumPy array containing the Cartesian product of the input sets,
#     where each row represents a unique combination.
#   """
#   n = sides.shape[0]  # Number of sets
#   F = np.meshgrid(*sides)  # Generate the ndgrid
#   G = np.zeros((np.prod([len(s) for s in sides]), n))  # Initialize output array
#   for i in range(n):
#     G[:, i] = F[i].flatten()  # Flatten and assign to output array

#   C = np.unique(G, axis=0)  # Find unique rows
#   return C

# # # example
# # # Define the sets
# # sides = np.array([
# #     [1, 2],    # First set
# #     [10, 20],  # Second set
# #     [100, 200] # Third set
# # ])

# # # Generate the Cartesian product
# # C = cartesian(sides)

# # # Print the result
# # print(C)



# from pyuncertainnumber.pba.intervals.backcalc import additive_bcc

import itertools
import numpy as np
from numbers import Number
from .params import Params


# there is an new `convert` func
def convert(un):
    """transform the input un into a Pbox object

    note:
        - theorically 'un' can be {Interval, DempsterShafer, Distribution, float, int}
    """

    from .pbox_base import Pbox
    from .dss import DempsterShafer
    from .distributions import Distribution

    if isinstance(un, Pbox):
        return un
    elif isinstance(un, Distribution):
        return un.to_pbox()
    elif isinstance(un, DempsterShafer):
        return un.to_pbox()
    else:
        raise TypeError(f"Unable to convert {type(un)} object to Pbox")


def interval_monte_carlo(f, x, y):
    pass


def p_backcalc(a, c, ops):
    """backcal for p-boxes
    #! incorrect implementation
    args:
        a, c (Pbox):probability box objects
        ops (object) : {'additive_bcc', 'multiplicative_bcc'} whether additive or multiplicative
    """
    from pyuncertainnumber.pba.intervals.intervalOperators import make_vec_interval
    from pyuncertainnumber.pba.aggregation import stacking
    from .pbox_base import Pbox
    from .intervals.number import Interval as I
    from .params import Params

    a_vs = a.to_interval()

    if isinstance(c, Pbox):
        c_vs = c.to_interval()
    elif isinstance(c, Number):
        c_vs = [I(c, c)] * Params.steps

    container = []
    for _item in itertools.product(a_vs, c_vs):
        container.append(ops(*_item))
    # print(len(container))  # shall be 40_000  # checkedout
    arr_interval = make_vec_interval(container)
    return stacking(arr_interval)


def adec(a, c):
    """
    Additive deconvolution: returns b such that a + b ≈ c
    Assumes a, b, c are instances of RandomNbr.

    note:
        implmentation from Scott
    """
    from .intervals.number import Interval as I
    from .pbox_abc import convert_pbox, Staircase

    n = Params.steps
    b = np.zeros(n)  # left bound of B, as in previous b.u[i]
    r = np.zeros(n)
    m = n - 1

    b[0] = c.left[0] - a.left[0]

    for i in range(1, m + 1):
        done = False
        sofar = c.left[i]
        for j in range(i):
            if sofar <= a.left[i - j] + b[j]:
                done = True
        if done:
            b[i] = b[i - 1]
        else:
            b[i] = c.left[i] - a.left[0]

    r[m] = c.right[m] - a.right[m]

    for i in range(m - 1, -1, -1):
        done = False
        sofar = c.right[i]
        for j in range(m, i, -1):
            if sofar >= a.right[i - j + m] + r[j]:
                done = True
        if done:
            r[i] = r[i + 1]
        else:
            r[i] = c.right[i] - a.right[m]

    # Check that bounds do not cross
    bad = any(b[i] > r[i] for i in range(n))

    if bad:
        # Try alternate method
        x = float("inf")
        y = float("-inf")
        for i in range(n):
            y = max(y, c.left[i] - a.left[i])
            x = min(x, c.right[i] - a.right[i])
        B = convert_pbox(I(y, x))
        return B

    # Final bounds check
    for i in range(n):
        if b[i] > r[i]:
            raise ValueError("Math Problem: couldn't deconvolve")
    return Staircase(left=b, right=r)


# * --------------- arithmetic with plain numbers --------------- *#



import numpy as np
import tqdm
from typing import Callable, Union
from scipy.optimize import brentq
from pyuncertainnumber.propagation.utils import Propagation_results


def cauchydeviates_method(
    x: np.ndarray,
    f: Callable,
    results: Propagation_results = None,
    n_sam: int = 500,
    xtol: Union[float, np.ndarray]  = 2e-12,
    rtol: Union[float, np.ndarray]  = 8.881784197001252e-16,
    maxiter: Union[float, np.ndarray]  = 100,
    min_y0: Union[float, np.ndarray] = 1e-6,   
    save_raw_data="no",
) -> Propagation_results:  # Specify return type
    """This method propagates intervals through a balck box model with the Cauchy deviate method. 
    It is an approximate method, so the user should expect non-identical results for different runs.

    args:
        x (np.ndarray): A 2D NumPy array representing the intervals for each input variable.
                         Each row should contain two elements: the lower and upper bounds of the interval.
        f (Callable): A callable function that takes a 1D NumPy array of input values
                        and returns a single output value or an array of output values.
                        Can be None, in which case only the Cauchy deviates (x) and the
                        maximum Cauchy deviate (K) are returned.
        n_sam (int): The number of samples (Cauchy deviates) to generate for each input variable (defaults 500 samples).
        results (Propagation_results): The class to use for storing results (defaults to Propagation_results).
        xtol (float): The absolute tolerance(s) for the root-finding algorithm (brentq). 
                      If a float, the same xtol is used for all outputs.If an np.ndarray, it should have the same length 
                      as the output of f, providing an absolute tolerance for brentq for each output. Default: 2e-12.
        rtol (float): The relative tolerance(s) for the root-finding algorithm (brentq). 
                      If a float, the same rtol is used for all outputs.If an np.ndarray, it should have the same length 
                      as the output of f, providing a relative tolerance for brentq for each output. Default: 8.88e-16.
        maxiter (int): The maximum number(s) of iterations for the root-finding algorithm (brentq). 
                       If a float, the same maxiter is used for all outputs.If an np.ndarray, it should have the same length 
                       as the output of f, providing a maximum number of iterations for brentq for each output. Default: 100.
        min_y0 (float, or np.ndarray): The lower bound(s) for the interval(s) in the brentq root-finding algorithm.
                                       If a float, the same lower bound is used for all outputs.
                                       If an np.ndarray, it should have the same length as the output of f,
                                       providing a separate lower bound for each output. Default: 1e-6.
        save_raw_data (str, optional): Acts as a switch to enable or disable the storage of raw input data when a function (f) 
          is not provided.
          - 'no': Returns an error that no function is provided.
          - 'yes': Returns the full arrays of unique input combinations.

    signature:
        cauchydeviate_method(x: np.ndarray, f: Callable, results_class = Propagation_results,
                        n_sam: int, save_raw_data='no') -> Propagation_results

    notes:
        The interval for the `brentq` root-finding algorithm is chosen heuristically.
        The lower bound is set to 1e-6, assuming the root (representing the change
        in the function's value) is positive and not exactly zero.  The upper bound
        is calculated as half the maximum absolute difference observed between the
        function's value at the central point and the values at the Cauchy deviate
        samples (`max(filtered_deltaF) / 2` or `max(filtered_deltaF_i) / 2`).
        This choice is based on the assumption that the actual change in the 
        function's value is likely to be smaller than the largest observed change.

        Users should be aware that these choices are heuristics and might need to be
        adjusted depending on the specific function and problem being solved. If
        `brentq` fails to converge, consider experimenting with different interval
        bounds.  For example, if the function is highly sensitive to input changes,
        the upper bound might be too large, and dividing by a larger number (e.g.,
        4 or 8 instead of 2) might be necessary.  Alternatively, if the root is
        expected to be very close to zero, the lower bound might need to be
        reduced.

    returns:
        A PropagationResult object containing the results.
            - 'un': UncertainNumber object(s) representing the uncertainty in the output(s) of the function.
            - 'raw_data': A dictionary containing raw data (if f is None):
                - 'x': Cauchy deviates (x).
                - 'K': Maximum Cauchy deviate (K).
                - 'bounds': An np.ndarray of the bounds for each output parameter (if f is not None).
                - 'min': A dictionary for lower bound results (if f is not None).
                    - 'f': Minimum output value(s).
                    - 'x': None (as input values corresponding to min/max are not tracked in this method).
                - 'max':  A dictionary for upper bound results (if f is not None).
                    - 'f': Maximum output value(s).
                    - 'x': None (as input values corresponding to min/max are not tracked in this method).

    raises:
        ValueError if no function is provided and save_raw_data is 'no'.

    example:
        >>> f = lambda x: x[0] + x[1] + x[2]  # Example function
        >>> x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
        >>> y = cauchydeviates_method(x_bounds,f=f, n_sam=50, save_raw_data = 'yes')
    """
    print(
        f"Total number of input combinations for the endpoints Cauchy deviates method: {n_sam}"
    )

    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    x = np.atleast_2d(x)  # Ensure x is 2D
    lo, hi = x.T  # Unpack lower and upper bounds directly

    xtilde = (lo + hi) / 2
    Delta = (hi - lo) / 2

    if f is not None:  # Only evaluate if f is provided
        ytilde = f(xtilde)

        if isinstance(ytilde, (float, np.floating)):  # Handle scalar output
            deltaF = np.zeros(n_sam)
            for k in tqdm.tqdm(range(1, n_sam), desc="Calculating Cauchy deviates"):
                r = np.random.rand(x.shape[0])
                c = np.tan(np.pi * (r - 0.5))
                K = np.max(c)
                delta = Delta * c / K
                x_sample = xtilde - delta
                deltaF[k] = K * (ytilde - f(x_sample))

            def Z(Del):
                return n_sam / 2 - np.sum(1 / (1 + (deltaF / Del) ** 2))
            
            mask = np.isnan(deltaF)
            filtered_deltaF = deltaF[~mask]

            # Use a small value for the lower bound
            zRoot = brentq(Z, min_y0, max(filtered_deltaF) / 2, 
                                              xtol =  xtol, rtol = rtol,maxiter=maxiter )
            min_candidate = np.array([ytilde - zRoot])
            max_candidate = np.array([ytilde + zRoot])
            bounds = np.array([min_candidate[0], max_candidate[0]])

            results.raw_data["min"] = np.array([{"x": None, "f": min_candidate}])
            results.raw_data["max"] = np.array([{"x": None, "f": max_candidate}])
            results.raw_data["bounds"] = bounds

        else:  # Handle array output
            len_y = len(ytilde)
            deltaF = np.zeros((n_sam, len_y))
            min_candidate = np.empty(len_y)
            max_candidate = np.empty(len_y)

            if isinstance(min_y0, float):
                min_y0_arr = np.full(len_y, min_y0)
            else:
                min_y0_arr = min_y0
            
            if isinstance(xtol, float):
                xtol_arr = np.full(len_y, xtol)
            else:
                xtol_arr = xtol
            
            if isinstance(rtol, float):
                rtol_arr = np.full(len_y, rtol)
            else:
                rtol_arr = rtol
            
            if isinstance(maxiter, float):
                maxiter_arr = np.full(len_y, maxiter)
            else:
                maxiter_arr = maxiter


            for k in tqdm.tqdm(range(n_sam), desc="Calculating Cauchy deviates"):
                r = np.random.rand(x.shape[0])
                c = np.tan(np.pi * (r - 0.5))
                K = np.max(c)
                delta = Delta * c / K
                x_sample = xtilde - delta
                result = f(x_sample)
                for i in range(len_y):
                    deltaF[k, i] = K * (ytilde[i] - result[i])

            bounds = np.zeros((len_y, 2))  # Initialize bounds array
            for i in range(len_y):
                mask = np.isnan(deltaF[:, i])
                filtered_deltaF_i = deltaF[:, i][~mask]

                def Z(Del):
                    return n_sam / 2 - np.sum(1 / (1 + (filtered_deltaF_i / Del) ** 2))

                try:  # Handle potential errors in brentq
                    # Use a small value for the lower bound
                    zRoot = brentq(Z, min_y0_arr[i], max(filtered_deltaF_i) / 2, 
                                              xtol =  xtol_arr[i], rtol = rtol_arr[i],maxiter=maxiter_arr[i] )
                except ValueError:
                    print(f"Warning: brentq failed for output {i}. Using 0 for zRoot.")
                    zRoot = 0  # Or handle the error in another way
                min_candidate[i] = ytilde[i] - zRoot
                max_candidate[i] = ytilde[i] + zRoot
                bounds[i] = [min_candidate[i], max_candidate[i]]

            # Populate the results object
            results.raw_data["bounds"] = bounds
            results.raw_data["min"] = np.append(
                results.raw_data["min"],
                [{"x": None, "f": min_val} for min_val in min_candidate],
            )
            results.raw_data["max"] = np.append(
                results.raw_data["max"],
                [{"x": None, "f": max_val} for max_val in max_candidate],
            )

        print("Input x for min max y are NOT available for the Cauchy method!")

        if save_raw_data == "yes":
            print("Input-Output raw data are NOT available for the Cauchy method!")

    elif save_raw_data == "yes":  # If f is None and save_raw_data is 'yes'
        x_samples = np.zeros((n_sam, x.shape[0]))
        K_values = np.zeros(n_sam)
        for k in tqdm.tqdm(range(n_sam), desc="Calculating Cauchy deviates"):
            r = np.random.rand(x.shape[0])
            c = np.tan(np.pi * (r - 0.5))
            K = np.max(c)
            delta = Delta * c / K
            x_samples[k] = xtilde - delta
            K_values[k] = K

        results.add_raw_data(x=x_samples, K=K_values)

    else:
        print(
            "No function is provided. Select save_raw_data = 'yes' to save the input combinations"
        )

    return results


# # # Example usage with different parameters for minimization and maximization
# f = lambda x: x[0] + x[1] + x[2]  # Example function

# # Determine input parameters for function and method
# x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
# n=50
# # Call the method
# y = cauchydeviates_method(x_bounds,f=f, n_sam=n, save_raw_data= 'yes')
# # # print the results
# # y.summary()
